-- ***************************************************************************
-- *  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de) 
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  02.08.2017  mifi  First version.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

entity uart_core is
   generic (
      FPGA_SYS_CLK_HZ : integer := 100000000;
      BAUDRATE        : integer := 115200
   );
   port (
      -- System
		clk_i       : in  std_logic                    := '0';
		rst_i       : in  std_logic                    := '0';

      -- Avalon
      address_i   : in  std_logic_vector(2 downto 0) := (others => '0');
      read_i      : in  std_logic                    := '0';
      readdata_o  : out std_logic_vector(7 downto 0);
      write_i     : in  std_logic                    := '0';
      writedata_i : in  std_logic_vector(7 downto 0) := (others => '0');
      
      -- Conduit
      txd_o       : out std_logic;
      rxd_i       : in  std_logic
   );
end entity uart_core;

-- ***************************************************************************
-- *  DEFINE: Architecture                                                   *
-- ***************************************************************************

architecture syn of uart_core is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   
   constant ADR_DATA   : std_logic_vector(2 downto 0) := "000"; -- Read/Write
   constant ADR_FRXU   : std_logic_vector(2 downto 0) := "001"; -- Read/Write
   constant ADR_FTXU   : std_logic_vector(2 downto 0) := "010"; -- Read/Write
   constant ADR_STATUS : std_logic_vector(2 downto 0) := "011"; -- Read
   
   constant ADR_RES_4 : std_logic_vector(2 downto 0) := "100"; -- Reserve 4
   constant ADR_RES_5 : std_logic_vector(2 downto 0) := "101"; -- Reserve 5
   constant ADR_RES_6 : std_logic_vector(2 downto 0) := "110"; -- Reserve 6
   constant ADR_RES_7 : std_logic_vector(2 downto 0) := "111"; -- Reserve 7
   

   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   component uart_rx is -- UART Receiver
      generic (
         FPGA_SYS_CLK_HZ : integer;
         BAUDRATE        : integer
      );
      port (
         -- System
         clk_i        : in  std_logic := '0';
         rst_i        : in  std_logic := '0';
      
         -- Uart line
         rxd_i        : in  std_logic := '0';
      
         -- FIFO
         fifo_data_o  : out std_logic_vector(7 downto 0);
         fifo_rd_i    : in  std_logic := '0';
         fifo_clr_i   : in  std_logic := '0';
         fifo_empty_o : out std_logic;
         fifo_full_o  : out std_logic;
         fifo_usedw_o : out std_logic_vector (7 downto 0);
      
         -- Debug
         debug_o      : out std_logic_vector(3 downto 0)
      );
   end component uart_rx;
   
   
   component uart_tx is -- UART Transmitter
      generic (
         FPGA_SYS_CLK_HZ : integer;
         BAUDRATE        : integer
      );
      port (
         -- System
         clk_i        : in  std_logic := '0';
         rst_i        : in  std_logic := '0';
      
         -- Uart line
         txd_o        : out std_logic := '0';
      
         -- FIFO
         fifo_data_i  : in  std_logic_vector(7 downto 0) := (others => '0');
         fifo_wr_i    : in  std_logic := '0';
         fifo_clr_i   : in  std_logic := '0';
         fifo_empty_o : out std_logic;
         fifo_full_o  : out std_logic;
         fifo_usedw_o : out std_logic_vector (7 downto 0)
      );
   end component uart_tx;
   
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   -- Register File
   signal rx_data       : std_logic_vector(7 downto 0) := (others => '0');
   signal tx_data       : std_logic_vector(7 downto 0) := (others => '0');
   signal status        : std_logic_vector(7 downto 0) := (others => '0');

   signal readdata      : std_logic_vector(7 downto 0) := (others => '0');
   signal read_mux      : std_logic_vector(7 downto 0) := (others => '0');
   
   -- RX-FIFO
   signal rx_fifo_clr   : std_logic := '0';
   signal rx_fifo_empty : std_logic := '0';
   signal rx_fifo_rd    : std_logic := '0';
   signal rx_fifo_rd_sr : std_logic_vector(1 downto 0) := (others => '0');
   signal rx_fifo_usedw : std_logic_vector(7 downto 0) := (others => '0');

   -- TX-FIFO
   signal tx_fifo_clr   : std_logic := '0';
   signal tx_fifo_full  : std_logic := '0';
   signal tx_fifo_wr    : std_logic := '0';
   signal tx_fifo_wr_sr : std_logic_vector(1 downto 0) := (others => '0');
   signal tx_fifo_usedw : std_logic_vector(7 downto 0) := (others => '0');

   signal rxd_sr        : std_logic_vector(1 downto 0) := (others => '1');
   signal rxd_sync      : std_logic := '0';
   signal txd           : std_logic := '0';
   
   -- Status
   signal status_dr     : std_logic := '0';
   signal status_thre   : std_logic := '0';
   
begin

   --
   -- Syncronize RXD
   --
   process (clk_i)
   begin
      if rising_edge(clk_i) then
         rxd_sr <= rxd_sr(0) & rxd_i;
      end if;
   end process;   
   
   rxd_sync <= rxd_sr(1);


   --------------------------------------------------------
   -- UART receiver
   --------------------------------------------------------

   --
   -- Handle rx_fifo_rd
   --
   process (clk_i)
   begin
      if rising_edge(clk_i) then
      
         -- Default
         rx_fifo_rd <= '0';
      
         if ((read_i = '1') AND (ADR_DATA = address_i)) then
            rx_fifo_rd_sr <= rx_fifo_rd_sr(0) & '1';
         else   
            rx_fifo_rd_sr <= rx_fifo_rd_sr(0) & '0';
         end if;
         
         -- Check for falling edge
         if ((rx_fifo_rd_sr(1) = '1') AND (rx_fifo_rd_sr(0) = '0')) then
            rx_fifo_rd <= '1';
         end if;
         
      end if;
   end process;   
   
   --
   -- Receiver   
   --
   inst_uart_rx : uart_rx
      generic map (
         FPGA_SYS_CLK_HZ => FPGA_SYS_CLK_HZ,
         BAUDRATE        => BAUDRATE
      )
      port map (
         -- System
         clk_i        => clk_i,
         rst_i        => rst_i,
         
         -- Uart line
         rxd_i        => rxd_sync,
      
         -- FIFO
         fifo_data_o  => rx_data,
         fifo_rd_i    => rx_fifo_rd,
         fifo_clr_i   => rx_fifo_clr,
         fifo_empty_o => rx_fifo_empty,
         fifo_full_o  => open,
         fifo_usedw_o => rx_fifo_usedw,
      
         -- Debug
         debug_o      => open
      );
      
   status_dr <= NOT rx_fifo_empty;      
   
   
   --------------------------------------------------------
   -- UART transmitter
   --------------------------------------------------------
   
   --
   -- Handle tx_fifo_wr
   --
   process (clk_i)
   begin
      if rising_edge(clk_i) then
      
         -- Default
         tx_fifo_wr <= '0';
      
         if ((write_i = '1') AND (ADR_DATA = address_i)) then
            tx_fifo_wr_sr <= tx_fifo_wr_sr(0) & '1';
         else   
            tx_fifo_wr_sr <= tx_fifo_wr_sr(0) & '0';
         end if;
         
         -- Check for falling edge
         if ((tx_fifo_wr_sr(1) = '1') AND (tx_fifo_wr_sr(0) = '0')) then
            tx_fifo_wr <= '1';
         end if;
         
      end if;
   end process;
   
   --
   -- Transmitter
   --
   inst_uart_tx : uart_tx
      generic map (
         FPGA_SYS_CLK_HZ => FPGA_SYS_CLK_HZ,
         BAUDRATE        => BAUDRATE
      )
      port map (
         -- System
         clk_i        => clk_i,
         rst_i        => rst_i,
      
         -- Uart line
         txd_o        => txd,
      
         -- FIFO
         fifo_data_i  => tx_data,
         fifo_wr_i    => tx_fifo_wr,
         fifo_clr_i   => tx_fifo_clr,
         fifo_empty_o => open,
         fifo_full_o  => tx_fifo_full,
         fifo_usedw_o => tx_fifo_usedw
      );
   
   status_thre <= NOT tx_fifo_full;
    
   
   --------------------------------------------------------
   -- Register File
   --------------------------------------------------------
   process (clk_i)
   begin
      if rising_edge(clk_i) then
      
         -- Default
         rx_fifo_clr <= '0';
         tx_fifo_clr <= '0';
         
      
         if (write_i = '1') then
         
            -- Address decoder
            case address_i is
            
               when ADR_DATA =>
                  tx_data <= writedata_i;
            
               when ADR_FRXU => 
                  rx_fifo_clr <= '1';
                  
               when ADR_FTXU => 
                  tx_fifo_clr <= '1';
                  
               when others =>
                  null; -- Do nothing
            
            end case;
         end if;
         
      end if; 
   end process;
   
   status <= "000000"    & -- Bit 7..2
             status_thre & -- Bit 1 
             status_dr;    -- Bit 0   
               
              
   --
   -- Read Multiplexer
   --
   read_mux <= rx_data       when (ADR_DATA   = address_i) else 
               rx_fifo_usedw when (ADR_FRXU   = address_i) else 
               tx_fifo_usedw when (ADR_FTXU   = address_i) else 
               status        when (ADR_STATUS = address_i) else (others => '0');

   process (clk_i)
   begin
      if rising_edge(clk_i) then
         if (read_i = '1') then
            readdata <= read_mux;
         end if;
      end if;
   end process;      

   
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

   -- Avalon
   readdata_o <= readdata;
   
   -- Conduit
   txd_o <= txd;

end architecture syn;

-- *** EOF ***
