-- ***************************************************************************
-- *  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de)
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  23.07.2017  mifi  First version.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.NUMERIC_STD.all;

-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

entity uart_rx is -- UART Receiver
   generic (
      FPGA_SYS_CLK_HZ : integer := 100000000;
      BAUDRATE        : integer := 115200
   );
   port (
      -- System
      clk_i        : in  std_logic := '0';
      rst_i        : in  std_logic := '0';
      
      -- Uart line
      rxd_i        : in  std_logic := '0';
      
      -- FIFO
      fifo_data_o  : out std_logic_vector(7 downto 0);
      fifo_rd_i    : in  std_logic := '0';
      fifo_clr_i   : in  std_logic := '0';
      fifo_empty_o : out std_logic;
      fifo_full_o  : out std_logic;
      fifo_usedw_o : out std_logic_vector (7 downto 0);
      
      -- Debug
      debug_o      : out std_logic_vector(3 downto 0)
   );
end entity uart_rx;


-- ***************************************************************************
-- *  DEFINE: Architecture                                                   *
-- ***************************************************************************

architecture syn of uart_rx is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   
   constant CLK_DIV    : integer := (FPGA_SYS_CLK_HZ/BAUDRATE) - 1;
   constant CLK_DIV_12 : integer := CLK_DIV / 2;
   constant CLK_DIV_34 : integer := (3*CLK_DIV) / 4;
   
   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   component fifo
      port (
         clock  : in std_logic ;
         data   : in std_logic_vector (7 downto 0);
         rdreq  : in std_logic ;
         sclr   : in std_logic ;
         wrreq  : in std_logic ;
         empty  : out std_logic ;
         full   : out std_logic ;
         q      : out std_logic_vector (7 downto 0);
         usedw  : out std_logic_vector (7 downto 0)
      );
   end component;
   
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------
   
   -- Receiver states
   type   fsm_rx_state_t is (FSM_STATE_IDLE, FSM_STATE_START, FSM_STATE_DATA, FSM_STATE_STOP);
   signal fsm_rx_state    : fsm_rx_state_t := FSM_STATE_IDLE;
   
   -- Shift registers
   signal noise_sr  : std_logic_vector(2 downto 0)  := (others => '0');
   signal rxd_sr    : std_logic_vector(1 downto 0)  := (others => '0'); 
   
   -- Misc
   signal cnt       : integer range 0 to CLK_DIV;
   signal rxd       : std_logic := '0'; 
   signal start_bit : std_logic := '0'; 
   signal bit_cnt   : integer range 0 to 15 := 0;
   signal data      : std_logic_vector(7 downto 0);
   signal data_wr   : std_logic := '0';   

   signal sample    : std_logic := '0'; 
   
begin

   --
   -- Noise filter (3 bit majority detector)
   --  
   process (clk_i)
   begin
      if rising_edge(clk_i) then
         noise_sr <= noise_sr(1 downto 0) & rxd_i;

         case (noise_sr) is
            when "000"  => rxd <= '0';
            when "001"  => rxd <= '0';
            when "010"  => rxd <= '0';
            when "011"  => rxd <= '1';
            when "100"  => rxd <= '0';
            when "101"  => rxd <= '1';
            when "110"  => rxd <= '1';
            when "111"  => rxd <= '1';
            when others => null;
         end case;
         
         rxd_sr <= rxd_sr(0) & rxd;
         
      end if;
   end process;
   
   
   --
   -- Find start bit
   --
   process (clk_i)
   begin
      if rising_edge(clk_i) then
      
         -- Check for falling edge
         if ((rxd_sr(1) = '1') AND (rxd_sr(0) = '0')) then
            start_bit <= '1';
         else
            start_bit <= '0';
         end if;
         
      end if;
   end process;
   
   
   --
   -- FSM
   --
   process (clk_i)
   begin
      if rising_edge(clk_i) then

         -- Sample clock
         if (cnt < CLK_DIV) then
            cnt <= cnt + 1;
         else
            cnt <= 0;
         end if;   
         
         -- Default
         data_wr <= '0';
         
         case (fsm_rx_state) is
         
            --
            -- FSM_STATE_IDLE
            --
            when FSM_STATE_IDLE =>
               cnt     <= 0;
               bit_cnt <= 0;
               sample  <= '0';
               
               if (start_bit = '1') then
                  fsm_rx_state <= FSM_STATE_START;
               end if;
            -- FSM_STATE_IDLE

            
            --
            -- FSM_STATE_START
            --
            when FSM_STATE_START =>
               if (cnt = CLK_DIV_12) then
                  sample <= NOT sample;
                  
                  -- Check if this is a start bit, signal must be 0
                  if (rxd_sr(1) = '1') then
                     -- This is not the start bit
                     fsm_rx_state <= FSM_STATE_IDLE;
                  else   
                     fsm_rx_state <= FSM_STATE_DATA;
                  end if;
               end if;
            -- FSM_STATE_START


            --
            -- FSM_STATE_DATA
            --
            when FSM_STATE_DATA =>
               if (bit_cnt < 8) then
                  if (cnt = CLK_DIV_12) then
                     sample <= NOT sample;
               
                     data(bit_cnt) <= rxd_sr(1);
                     bit_cnt       <= bit_cnt + 1;
                  end if;
               else    
                  fsm_rx_state <= FSM_STATE_STOP;
               end if;                  
            -- FSM_STATE_DATA
            
            
            --
            -- FSM_STATE_STOP
            --
            when FSM_STATE_STOP =>
               if (cnt = CLK_DIV_12) then
                  sample  <= NOT sample;
                  bit_cnt <= bit_cnt + 1;
               end if;   
               
               if ((bit_cnt = 9) AND (cnt = CLK_DIV_34)) then
                  data_wr      <= '1';
                  fsm_rx_state <= FSM_STATE_IDLE;
               end if;
            -- FSM_STATE_STOP


            when others =>
               fsm_rx_state <= FSM_STATE_IDLE;

         end case;
      
      end if;
   end process;
   
   
   --
   -- RX FIFO
   --
   inst_rx_fifo : fifo
      port map (
         clock  => clk_i,
         
         data   => data,
         wrreq  => data_wr,

         q      => fifo_data_o,
         rdreq  => fifo_rd_i,
         sclr   => fifo_clr_i,
         empty  => fifo_empty_o,
         full   => fifo_full_o,
         usedw  => fifo_usedw_o
      );

      
   --------------------------------------------------------
   -- Output
   --------------------------------------------------------
   
   -- Debug   
   debug_o(0) <= rxd_sr(1);
   debug_o(1) <= sample;
   debug_o(2) <= '0';
   debug_o(3) <= '0';

end architecture syn;

-- *** EOF ***

