-- ***************************************************************************
-- *  Copyright (c) 2017 by Michael Fischer (www.emb4fun.de) 
-- *  All rights reserved.
-- *
-- *  Redistribution and use in source and binary forms, with or without 
-- *  modification, are permitted provided that the following conditions 
-- *  are met:
-- *  
-- *  1. Redistributions of source code must retain the above copyright 
-- *     notice, this list of conditions and the following disclaimer.
-- *  2. Redistributions in binary form must reproduce the above copyright
-- *     notice, this list of conditions and the following disclaimer in the 
-- *     documentation and/or other materials provided with the distribution.
-- *  3. Neither the name of the author nor the names of its contributors may 
-- *     be used to endorse or promote products derived from this software 
-- *     without specific prior written permission.
-- *
-- *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
-- *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
-- *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
-- *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
-- *  THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
-- *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
-- *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
-- *  OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
-- *  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
-- *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF 
-- *  THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
-- *  SUCH DAMAGE.
-- *
-- ***************************************************************************
-- *  History:
-- *
-- *  12.08.2017  mifi  First version.
-- ***************************************************************************

-- ***************************************************************************
-- *  DEFINE: Library                                                        *
-- ***************************************************************************
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;


-- ***************************************************************************
-- *  DEFINE: Entity                                                         *
-- ***************************************************************************

entity top is
   port ( 
      -- Input clock 
      CLOCK_50      : in  std_logic;
      
      --
      -- SDRAM interface,
      -- here a A3V64S40ETP-G6 (166MHz@CL-3) is used.
      -- Reference is made to Zentel datasheet:
      -- A3V64S40ETP, Revision 1.2, Mar., 2010
      --        
      SDRAM_CLK     : out   std_logic;                       -- Master Clock
      SDRAM_CKE     : out   std_logic;                       -- Clock Enable    
      SDRAM_CS_N    : out   std_logic;                       -- Chip Select
      SDRAM_RAS_N   : out   std_logic;                       -- Row Address Strobe
      SDRAM_CAS_N   : out   std_logic;                       -- Column Address Strobe
      SDRAM_WE_N    : out   std_logic;                       -- Write Enable
      SDRAM_DQ      : inout std_logic_vector(15 downto 0);   -- Data I/O (16 bits)
      SDRAM_DQML    : out   std_logic;                       -- Output Disable / Write Mask (low)
      SDRAM_DQMU    : out   std_logic;                       -- Output Disable / Write Mask (high)
      SDRAM_ADDR    : out   std_logic_vector(12 downto 0);   -- Address Input (12 bits)
      SDRAM_BA_0    : out   std_logic;                       -- Bank Address 0
      SDRAM_BA_1    : out   std_logic;                       -- Bank Address 1
      
      --
      -- LEDs, green and heartbeat
      --
      LED           : out   std_logic_vector(6 downto 0);
      LED_HEARTBEAT : out   std_logic;
      
      --
      -- Swithes and keys
      --
      KEY           : in    std_logic_vector(1 downto 0);
      SW            : in    std_logic_vector(3 downto 0);
      
      --
      -- EPCS
      --
      EPCS_DCLK     : out   std_logic;
      EPCS_NCSO     : out   std_logic;
      EPCS_ASDO     : out   std_logic;
      EPCS_DATA0    : in    std_logic;
      
      --      
      -- UART
      --
      UART_RXD      : in    std_logic;
      UART_TXD      : out   std_logic
   );
end entity top;


-- ***************************************************************************
-- *  DEFINE: Architecture                                                   *
-- ***************************************************************************

architecture syn of top is

   --------------------------------------------------------
   -- Define all constants here
   --------------------------------------------------------
   
   --------------------------------------------------------
   -- Define all components which are included here
   --------------------------------------------------------
   
   component pll_sys
      port(
         inclk0 : in  std_logic := '0';
         c0     : out std_logic;
         c1     : out std_logic;
         c2     : out std_logic;
         locked : out std_logic 
      );
   end component pll_sys;


   component heartbeat is
      generic (
         SYS_CLK_HZ : integer;
         PERIOD_MS  : integer
      );
      port (  
         clk_i : in  std_logic := '0';
         led_o : out std_logic
      );
   end component heartbeat;
   
   
   component qsys_core is
      port (
         clk_clk           : in    std_logic                     := 'X';             -- clk
         reset_reset_n     : in    std_logic                     := 'X';             -- reset_n

         clk_rsu_clk       : in    std_logic                     := 'X';             -- clk
         reset_rsu_reset_n : in    std_logic                     := 'X';             -- reset_n
         
         sdram_addr        : out   std_logic_vector(12 downto 0);                    -- addr
         sdram_ba          : out   std_logic_vector(1 downto 0);                     -- ba
         sdram_cas_n       : out   std_logic;                                        -- cas_n
         sdram_cke         : out   std_logic;                                        -- cke
         sdram_cs_n        : out   std_logic;                                        -- cs_n
         sdram_dq          : inout std_logic_vector(15 downto 0) := (others => 'X'); -- dq
         sdram_dqm         : out   std_logic_vector(1 downto 0);                     -- dqm
         sdram_ras_n       : out   std_logic;                                        -- ras_n
         sdram_we_n        : out   std_logic;                                        -- we_n
         
         pio_led_export    : out   std_logic_vector(6 downto 0);                     -- export
         pio_key_export    : in    std_logic_vector(1 downto 0)  := (others => 'X'); -- export
         pio_sw_export     : in    std_logic_vector(3 downto 0)  := (others => 'X'); -- export
         
         epcs_dclk         : out   std_logic;                                        -- dclk
         epcs_sce          : out   std_logic;                                        -- sce
         epcs_sdo          : out   std_logic;                                        -- sdo
         epcs_data0        : in    std_logic                     := 'X';             -- data0

         uart_txd_o        : out   std_logic;                                        -- txd_o
         uart_rxd_i        : in    std_logic                     := 'X'              -- rxd_i
      );
   end component qsys_core;
   
   
   --------------------------------------------------------
   -- Define all local signals here
   --------------------------------------------------------

   signal sys_clk       : std_logic := '0';
   signal clk_32        : std_logic := '0';
   signal pll_locked    : std_logic := '0';
   signal reset         : std_logic := '0';
   signal reset_s1      : std_logic := '1';
   signal reset_s2      : std_logic := '1';
   signal reset_s3      : std_logic := '1';
   signal sys_rst       : std_logic;

   signal sdram_ba      : std_logic_vector(1 downto 0);
   signal sdram_dqm     : std_logic_vector(1 downto 0);
   
begin

   --
   -- PLL
   --
   inst_pll_sys : pll_sys
      port map ( 
         inclk0 => CLOCK_50,
         c0     => sys_clk,
         c1     => SDRAM_CLK,
         c2     => clk_32,
         locked => pll_locked
      );


   --
   -- In general it is a bad idea to use an asynchhronous Reset signal.
   -- But it is only a bad idea in case of asynchhronous deasserting. 
   -- Therefore the deasserting of the Reset signal must be synchronized.
   --               
                              
   -- Asynchronous assert  
   reset <= '1' when (pll_locked = '0') else '0';

   -- Synchronize deassert
   process (sys_clk, reset)
   begin
      if (reset = '1') then
         reset_s1 <= '1';
         reset_s2 <= '1';
         reset_s3 <= '1';
      elsif rising_edge(sys_clk) then
         reset_s1 <= '0';
         reset_s2 <= reset_s1;
         reset_s3 <= reset_s2; 
      end if;
   end process;   

   -- The deassert edge is now synchronized   
   sys_rst <= reset_s3; 
      

   --
   -- Heartbeat
   --
   inst_heartbeat : heartbeat 
      generic map (
         SYS_CLK_HZ => 32000000,
         PERIOD_MS  => 1000
      )
      port map ( 
         clk_i => clk_32,
         led_o => LED_HEARTBEAT
      );
      
      
   --
   -- Qsys core
   --      
   inst_qsys_core : qsys_core
      port map (
         clk_clk           => sys_clk,
         reset_reset_n     => not sys_rst,
         
         clk_rsu_clk       => clk_32,
         reset_rsu_reset_n => not sys_rst,

         sdram_addr        => SDRAM_ADDR,
         sdram_ba          => sdram_ba,
         sdram_cas_n       => SDRAM_CAS_N,
         sdram_cke         => SDRAM_CKE,
         sdram_cs_n        => SDRAM_CS_N,
         sdram_dq          => SDRAM_DQ,
         sdram_dqm         => sdram_dqm,
         sdram_ras_n       => SDRAM_RAS_N,
         sdram_we_n        => SDRAM_WE_N,
         
         pio_led_export    => LED,
         pio_key_export    => KEY,
         pio_sw_export     => SW,
         
         epcs_dclk         => EPCS_DCLK,
         epcs_sce          => EPCS_NCSO,
         epcs_sdo          => EPCS_ASDO,
         epcs_data0        => EPCS_DATA0,
         
         uart_txd_o        => UART_TXD,
         uart_rxd_i        => UART_RXD
      );
            
   SDRAM_BA_1 <= sdram_ba(1);
   SDRAM_BA_0 <= sdram_ba(0);
   
   SDRAM_DQMU <= sdram_dqm(1);
   SDRAM_DQML <= sdram_dqm(0);  
   

   --------------------------------------------------------
   -- Output
   --------------------------------------------------------

end architecture syn;

-- *** EOF ***

